import pandas as pd
import numpy as np

from pandas.api.types import is_float_dtype, is_numeric_dtype, is_string_dtype
#from humailib.cloud_tools import GoogleBigQuery, GoogleCloudStorage

"""
Column type definitions from the data dictionary.
    
https://humai.sharepoint.com/:w:/s/Platform/ET6ZzN-QgiJIpLdbVMkFz0oBe9_UOhSoau0Zvbo7HJ-5lQ?e=OxMlYv
"""
humai_table_field_types = {    
    'transactions' : {
        'datetime' : ['Datetime'],
        'string' : ['Transaction_ID','Branch_ID','Customer_ID','Status','Single_Item_Product_ID'],
        'float' : ['Total_Gross','Discount','Amount_Paid'],
        'int' : ['Total_Num_Items_Purchased']
    },
    'items' : {
        'datetime' : ['Datetime'],
        'string' : ['Transaction_ID','Product_ID','Name'],
        'float' : ['Total', 'Discount'],
    },
    'mailings' : {
        'datetime' : ['Send_Datetime'],
        'string' : ['Broadcast_ID','Campaign_ID'],
        'int' : ['Num_Customers']
    },
    'mailings_events' : {
        'datetime' : ['Activity_Datetime', 'Send_Datetime'],
        'string' : ['Broadcast_ID','Customer_ID','Activity','Description'],
    },
    'customers' : {
        'string' : ['Customer_ID','Email','Name','Address'],
    },
    'products' : {
        'string' : ['Name','Product_ID','SKU'],
        'float' : ['Price'],
    },
    'customer_characteristics' : {
        'string' : ['Customer_ID']
    },
    'product_characteristics' : {
        'string' : ['Customer_ID']
    },
    'predictions' : {
        'string' : ['Customer_ID'],
        'datetime' : ['Date'],
        #'int' : ['Time_Of_Day']
    },
}


def convert_table_columns(
    df, 
    table_type = 'transactions', 
    datetime_format='%Y-%m-%d %H:%M:%S'
):
    
    """
    Convert table column types to table column type definitions as specified in the data dictionary:
    
    https://humai.sharepoint.com/:w:/s/Platform/ET6ZzN-QgiJIpLdbVMkFz0oBe9_UOhSoau0Zvbo7HJ-5lQ?e=OxMlYv
    """
    
    types = humai_table_field_types[table_type]
    if 'datetime' in types:
        columns_as_datetime(df, columns=types['datetime'], datetime_format=datetime_format)
    if 'string' in types:
        columns_as_str(df, columns=types['string'], uppercase=False, drop_empty=False)
    if 'float' in types:
        columns_as_float(df, columns=types['float'], na_replace_value=None, drop_na=False)
    if 'int' in types:
        columns_as_int(df, columns=types['int'], na_replace_value=None, drop_na=False)

    for c in df.columns: print("  {} -> {}".format(c, df[c].dtype))
        
        
def load_table(
    gbq, 
    dataset_table,
    table_type,
    datetime_columns = None,
    datetime_format = '%Y-%m-%d %H:%M:%S', 
    flush_cache = True,
):
    """
    Load table and convert column types
    """    
         
    print("[load_table] Loading table...")
    df = gbq.download_table_to_pandas(dataset_table, flush_cache=flush_cache)
    
    print("[load_table] Converting table columns...")
    if datetime_columns is not None:
        columns_as_datetime(df, columns=datetime_columns, datetime_format=datetime_format)
    convert_table_columns(df, table_type=table_type, datetime_format=datetime_format)    
    
    return df


def load_csv(
    gcs,
    filename,
    table_type,
    datetime_columns = None,
    datetime_format='%Y-%m-%d %H:%M:%S',
    flush_cache = True
):
    """
    Load CSV from GCS and convert column types.
    """
    
    print("[load_table] Loading table...")
    df = gcs.download_csv(filename, flush_cache=flush_cache)
    
    print("[load_table] Converting table columns...")
    if datetime_columns is not None:
        columns_as_datetime(df, columns=datetime_columns, datetime_format=datetime_format)
    convert_table_columns(df, table_type=table_type, datetime_format=datetime_format)
    
    return df


def load_and_merge_tables(
    gbq, 
    dataset,
    table_type,
    date_ranges,
    datetime_columns = None,
    datetime_format = '%Y-%m-%d %H:%M:%S', 
    flush_cache = True,
):
    """
    1. Load tables by date range. 
    2. Remove duplicate rows.
    3. Convert column types
    """    
    
    print("[load_and_merge_tables] Loading tables...")
    df = None
    for date_range in date_ranges:
        df_tmp = gbq.download_table_to_pandas(
            "{}.{}_{}".format(dataset, table_type, date_range),
            flush_cache=flush_cache
        )
        if df is None:
            df = df_tmp.copy()
        else:
            df = df.append(df_tmp)

    print("[load_and_merge_tables] Dropping duplicate rows...")
    df.drop_duplicates(inplace=True)
    
    print("[load_and_merge_tables] Converting table columns...")
    if datetime_columns is not None:
        columns_as_datetime(df, columns=datetime_columns, datetime_format=datetime_format)
    convert_table_columns(df, table_type=table_type, datetime_format=datetime_format)
    
    return df
    
def force_string(value):
    return 'S' + str(value)#'S' + str(value)

def column_stats(df, col_name, max_rows=5):
    """
    Print out column statistics.
    """
    
    isna = df[col_name].isna().sum()
    print("NaNs: {:,} out of {:,} ({:.2f}%)".format(isna, len(df), 100.0*isna/len(df)))
    print("== Describe ==")
    print(df[col_name].describe())
    
    if not is_float_dtype(df[col_name]):
        print("== Value counts ==")
        print(df[col_name].value_counts()[:max_rows])
        
    print("== Head ==")
    print(df[col_name].head(n=max_rows))
    
          
def columns_as_str(df, columns, uppercase=False, drop_empty=False):
          
    """
    Convert columns to string, and convert to uppercase if specified.

    Note: This is done inplace.
    """
    
    for c in columns:
        if c in df:
            if uppercase:    
                df.loc[:,c] = df[c].astype(str, skipna=True).str.upper()
            else:
                df.loc[:,c] = df[c].astype(str, skipna=True)

            if drop_empty:
                df.loc[:,c].replace(to_replace='', value=np.nan, inplace=True)
                df.dropna(subset=[c], inplace=True)
            else:
                df.loc[:,c].fillna('', inplace=True)

            
def columns_as_datetime(df, columns, datetime_format='%Y-%m-%d %H:%M:%S'):
    
    """
    Convert columns to time-zone-agnostic datetime.

    Note: This is done inplace.
    """
    
    for c in columns:
        if c in df:
            df.loc[:,c] = pd.to_datetime(df[c], format=datetime_format, utc=True).dt.tz_convert(None)


def columns_as_float(df, columns, na_replace_value=None, drop_na=True):
    
    """
    Convert columns to float.

    Note: This is done inplace.
    """
    
    if columns is None:
        columns = df.columns
        
    if na_replace_value is None:
        na_replace_value = np.nan
        
    for c in columns:
        if c in df and not is_float_dtype(df[c]):
            
            if is_string_dtype(df[c]):
                df.loc[:,c].replace({
                    '' : np.nan, 'nan' : np.nan, 'NaN' : np.nan, 'None' : np.nan
                }, inplace=True)
                
            if drop_na:
                df.dropna(subset=[c], how='any', inplace=True)
            else:
                df.loc[:,c].fillna(na_replace_value, inplace=True)
                
            df.loc[:,c] = df[c].astype('float64', skipna=False)
        

def columns_as_int(df, columns, na_replace_value=None, drop_na=True):
    
    """
    Convert string/float columns to integer.

    Note: This is done inplace.
    """
    
    if not drop_na and na_replace_value is None:
        na_replace_value = -1.0

    columns_as_float(df, columns, na_replace_value, drop_na)
    
    for c in columns:
        if c in df:
            df.loc[:,c] = df[c].astype('int64', skipna=False)


def transactions_cleanup(df, datetime_column, tid_column = 'Transaction_ID', cid_column = 'Customer_ID', datetime_format='%Y-%m-%d %H:%M:%S'):
    
    """
    Perform common transaction data cleanup.

    Note: This is done inplace.
    """
    
    df.loc[:,datetime_column] = pd.to_datetime( df[datetime_column], format=datetime_format)
    df.loc[:,datetime_column].dropna(inplace=True)

    # Get transactions date range
    print('Date range from {0} to {1}'.format(df[datetime_column].min(), df[datetime_column].max())) 
    
    # Remove column 'Unnamed: 0' if it exists
    if 'Unnamed: 0' in df:
        df.drop('Unnamed: 0', axis=1, inplace=True)
        print("Removed unnamed index column")

    print("Before dropping empty Transaction IDs: {:,}".format(len(df)))

    # Drop Transaction_ID's that are empty
    df.loc[:,tid_column].replace('', np.nan, inplace=True)
    df.dropna(subset=[tid_column], inplace=True)

    print("After: {:,}".format(len(df)))
    
    print("Before dropping empty Customer IDs: {:,}".format(len(df)))

    # Drop [cid]'s that are empty
    df.loc[:,cid_column].replace('', np.nan, inplace=True)
    df.dropna(subset=[cid_column], inplace=True)

    print("After: {:,}".format(len(df)))


def email_ensure_zero_or_one_activity_per_broadcast(
    df, cid_column = 'Customer_ID', activity_column = 'Activity', 
    activity_datetime_column = 'Activity_Datetime',
    verbose=True
):

    """
    Ensure each email sent to each customer has only one event: either a delivery, an open, a click, or a transaction,
    in that order. Keep only the first (in time) occurrence of it.

    Note: This is done inplace.
    """
    
    conv_table = {
        'DELIVERY' : 0,
        'OPEN' : 1,
        'CLICK' : 2,
        'TRANSACTION' : 3
    }
    
    if verbose:
        print("[email_ensure_zero_or_one_activity_per_broadcast] converting activities to ordinals...")
    
    #df.loc[:,'activity_id'] = df[activity_column].apply(
    #    lambda x: conv_table[x] if x in conv_table else -1
    #)
    df.loc[:,'activity_id'] = df[activity_column].replace(to_replace=conv_table)

    if verbose:
        print("[email_ensure_zero_or_one_activity_per_broadcast] get max...")

    result = df.groupby(by=[cid_column, 'Broadcast_ID'])['activity_id'].max()
    
    if verbose:
        print("[email_ensure_zero_or_one_activity_per_broadcast] determining which rows to keep...")
        
    df.loc[:,'keep'] = df.apply(lambda x: 1 if result[x[cid_column],x['Broadcast_ID']] == x['activity_id'] else 0, axis=1)

    if verbose:
        print("[email_ensure_zero_or_one_activity_per_broadcast] dropping...")
    df.drop(index=df[df.keep == 0].index, inplace=True)
    
    if verbose:
        print("[email_ensure_zero_or_one_activity_per_broadcast] sorting and keeping only the first occurence of the activity...")
        
    df.sort_values(by=[cid_column, 'Broadcast_ID', activity_datetime_column], inplace=True)
    df.drop_duplicates(subset=[cid_column, 'Broadcast_ID'], keep='first', inplace=True)

    del df['keep']
    del df['activity_id']



def email_keep_activities(df, activities):
    
    """
    Keep only certain email activities, remove the rest.

    Note: This is done inplace.
    """
    
    keep = df.Activity.isin(activities)
    df.drop(index=df[~keep].index, inplace=True)
    
    
def email_activity_stats(df, activity_column='Activity'):
    
    n = []
    n.append(len(df[df[activity_column] == 'DELIVERY']))
    n.append(len(df[df[activity_column] == 'OPEN']))
    n.append(len(df[df[activity_column] == 'CLICK']))
    n.append(len(df[df[activity_column] == 'TRANSACTION']))
    n.append(len(df[df[activity_column] == 'UNSUBSCRIBE']))
    n.append(len(df) - np.sum(n))
    N = np.sum(n)
    
    stats = {
        'Delivered' : n[0],
        'Delivered_perc' : 100.0 * n[0]/N,
        'Opens' : n[1],
        'Opens_perc' : 100.0 * n[1]/N,
        'Clicks' : n[2],
        'Clicks_perc' : 100.0 * n[2]/N,
        'Transactions' : n[3],
        'Transactions_perc' : 100.0 * n[3]/N,
        'Unsubscribes' : n[4],
        'Unsubscribes_perc' : 100.0 * n[4]/N,
        'Other' : n[5],
        'Other_perc' : 100.0 * n[5]/N,
        'Total' : np.sum(n)
    }

    return stats


def print_email_activity_stats(df, activity_column='Activity'):
    
    stats = email_activity_stats(df, activity_column=activity_column)
    
    print("Deliveries:   {:,} ({:.2f} %)".format(stats['Delivered'], stats['Delivered_perc']))
    print("Opens:        {:,} ({:.2f} %)".format(stats['Opens'], stats['Opens_perc']))
    print("Clicks:       {:,} ({:.2f} %)".format(stats['Clicks'], stats['Clicks_perc']))
    print("Transactions: {:,} ({:.2f} %)".format(stats['Transactions'], stats['Transactions_perc']))
    print("Unsubscribes: {:,} ({:.2f} %)".format(stats['Unsubscribes'], stats['Unsubscribes_perc']))
    print("Other:        {:,} ({:.2f} %)".format(stats['Other'], stats['Other_perc']))
    print("Total:        {:,}".format(stats['Total']))
    
    



import pandas as pd

import humailib.utils as hu
import humailib.hasher as hh
from humailib.cloud_tools import GoogleCloudStorage, GoogleBigQuery

class CustomerIDFromNameAddressEmail:
    
    def __init__(self):
        self.cid = 0
        self.customer_id_lookup = {}
        self.customer_table = []
        
    def generate(
        self, 
        df, name_col='Name', addr_col='Address', email_col='Email', 
        verbose=False
    ):
      
        num_new = 0
        ii = 1
        n = df[name_col].nunique()
        cids = {}
        
        dfg = df[[name_col, addr_col, email_col, 'Transaction_ID']].groupby(by=[name_col])
        
        if verbose:
            print("0/{:,} ({:.2f}%)".format(n, 0))
        
        for name, data in dfg:
            #print(name)
            
            name = name.strip().upper()
    
            if verbose and ii % 5000 == 0:
                print("{:,}/{:,} ({:.2f}%) New IDs created: {:,}".format(ii, n, 100.0*ii/n, num_new))

            id_by_email = {}
            id_by_addr = {}
            if name in self.customer_id_lookup:
                id_by_email, id_by_addr = self.customer_id_lookup[name]
    
            for index, row in data.iterrows():
                new_email = row[email_col].strip().upper()
                new_addr = row[addr_col].strip().upper()
                if not new_email in id_by_email and not new_addr in id_by_addr:
                    id_by_email[ new_email ] = self.cid
                    id_by_addr[ new_addr ] = self.cid
                    #print(cid)
                    #cids.append(self.cid)
                    cids[str(row['Transaction_ID'])] = self.cid
                    
                    self.customer_table.append([self.cid, name, new_email, new_addr])
                    self.cid = 1 + self.cid
                    num_new = 1 + num_new
                else:
                    if new_email in id_by_email:
                        id_by_addr[ new_addr ] = id_by_email[ new_email ]
                    elif new_addr in id_by_addr:
                        id_by_email[ new_email ] = id_by_addr[ new_addr ]

                    #print(addr[ ad ])
                    #cids.append( id_by_addr[ new_addr ] )
                    cids[str(row['Transaction_ID'])] = id_by_addr[new_addr]
                    
                    #if name == 'MIRANDA VEREGGEN':
                    #    print("Miranda: {}".format(id_by_addr[new_addr]))
                        
                #if self.cid == '91962' or id_by_addr[new_addr] == '91962':
                #    print("91962: {}".format(name))

            self.customer_id_lookup[name] = [id_by_email, id_by_addr]

            ii = 1 + ii
            
        if verbose:
            print("{:,}/{:,} ({:.2f}%). New IDs created: {:,}".format(n, n, 100.0, num_new))
    
        return cids
    
    def get_customer_id_table(self):
        
        if self.customer_table is not None:
            
            data = [tuple(c) for c in self.customer_table]
            df = pd.DataFrame(data, columns=['Customer_ID','Name','Email','Address'])
            hu.columns_as_str(df, columns=['Customer_ID', 'Name','Email','Address'])
            
            return df
        
        return None
    
    def upload_and_replace(self, dataset_table_name, gbq):
        
        if self.customer_table is None or self.customer_id_lookup is None:
            return
        
        if gbq is None:
            gbq = GoogleBigQuery()
        df = self.get_customer_id_table()

        hh.encrypt_columns(df, columns=['Name','Email','Address'])
        gbq.upload_and_replace_table(df, dataset_table_name)
        
    def merged_and_upload(self, dataset_table_name, gbq, flush_cache=True):
        
        if self.customer_table is None or self.customer_id_lookup is None:
            return
        
        if gbq is None:
            gbq = GoogleBigQuery()
        #df_old = gbq.download_table_to_pandas(dataset_table_name)
        df_old = hu.load_table(gbq, dataset_table_name, table_type='customers', flush_cache=flush_cache)
        if df_old is None:
            return self.upload_and_replace_customers_to_bq(dataset_table_name)
        
        # Keep only those rows that we haven't touched.
        df_old.drop(columns=['Name','Email','Address'], inplace=True)
        #hu.columns_as_str(df_old, columns=['Name','Email','Address'])
        #hh.decrypt_columns(df_old, columns=['Name','Email','Address'])
        
        print("df_old:")
        print(df_old.head())
        
        df_new = self.get_customer_id_table()
        
        print("df_new:")
        print(df_new.head())
        
        df_merged = df_old.merge(df_new, left_on='Customer_ID', right_on='Customer_ID', how='outer')
        
        print("df_merged:")
        print(df_merged.head())

        hh.encrypt_columns(df_merged, columns=['Name','Email','Address'])
        gbq.upload_and_replace_table(df_merged, dataset_table_name)
        
    def download(self, dataset_table_name, gbq, flush_cache=True):
        
        self.__init__()
        
        if gbq is None:
            gbq = GoogleBigQuery()
        #df = gbq.download_table_to_pandas(dataset_table_name, from_cache)
        df = hu.load_table(gbq, dataset_table_name, table_type='customers', flush_cache=flush_cache)
        if df is None:
            return False
        
        hh.decrypt_columns(df, columns=['Name','Email','Address'])
        
        self.cid = len(df)
        self.customer_table = df[['Customer_ID','Name','Email','Address']].to_numpy().tolist()
        
        for d in self.customer_table:
            
            cid = d[0]
            name = d[1]

            id_by_email = {}
            id_by_addr = {}
            if name in self.customer_id_lookup:
                id_by_email, id_by_addr = self.customer_id_lookup[name]
                
            new_email = d[2]
            new_addr = d[3]
            
            id_by_email[ new_email ] = cid
            id_by_addr[ new_addr ] = cid

            self.customer_id_lookup[name] = [id_by_email, id_by_addr]
            
        return True
        
        


import pandas as pd
import numpy as np
import datetime

from pathlib import Path

from sklearn.pipeline import Pipeline
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.externals import joblib

from humailib.cloud_tools import GoogleCloudStorage
from humailib.transformers import DataframeFeatureUnion

class PipelineBuilder:
    """
    This class allows you to build a pipeline, by executing and adding one ore more steps 
    at a time, in a way that lends itself well to how data transformation is implemented in Notebooks.
    """
    
    def __init__(self):
        
        self.steps = {}
        self.step_names = {}
        
        self.current = 'default'
        self.steps[self.current] = []
        self.step_names[self.current] = {}
        
        return
    
    def new_pipeline(self, name):
        
        self.steps[name] = []
        self.step_names[name] = {}
        self.current = name
    
    def transform_and_add(self, steps, df):
        
        pipeline = Pipeline(steps)
        df_out = pipeline.transform(df)
        
        self.add(steps)
        
        return df_out
    
    def add(self, steps):
        
        for step in steps:
            if step[0] not in self.step_names[self.current]:
                self.steps[self.current].append(step)
                self.step_names[self.current][step[0]] = len(self.steps[self.current])-1
                
        return
    
    def build(self, pipelines=None, catch_all_params=None, verbose=True):
        
        assert pipelines is not None and len(pipelines)>0
        
        if isinstance(pipelines, str):
            pipelines = [pipelines]
        
        sub_names = []
        sub_pipes = []
        for pipeline_name in pipelines:
            if pipeline_name in self.steps and len(self.steps[pipeline_name]) > 0:
                sub_names.append(pipeline_name)
                sub_pipes.append(self.steps[pipeline_name])
                
        pipe_out = Pipeline([
            (name, Pipeline(sub)) for name,sub in zip(sub_names,sub_pipes)
        ])
        
        if verbose:
            print("[PipelineBuilder::build] Built:")
            print(pipe_out.steps)
        
        if catch_all_params is not None:
            for catch_all_name, value in catch_all_params.items():
                for pipeline_name in pipelines:
                    # Get the parameters for all components in this pipeline.
                    components_params = pipe_out.named_steps[pipeline_name].get_params()
                    # See if any parameters have names that contain the catch-all name
                    params = {name:value for name,_ in components_params.items() if catch_all_name in name}
                    # Set these to have the catch-all value
                    if verbose:
                        print("[PipelineBuilder::build] Setting params {}".format(params))
                    pipe_out.named_steps[pipeline_name].set_params(**params)
        
        return pipe_out
    
    def load(self, filename, gcs=None, cache_dir='./cache'):
        
        if 'gs://' in filename:
            cache_file = Path(cache_dir + '/' + filename.split('/')[-1])
            gcs.download_file(filename, cache_file)
            pipeline = joblib.load(cache_file)
        else:
            pipeline = joblib.load(filename)
            
        self.__init__()
        for name, sub in pipeline.steps:
            self.steps[ name ] = sub.steps
            self.step_names[ name ] = {sub_step_name:i for i,(sub_step_name,_) in enumerate(sub.steps)}
            
        return
    
    def save(self, filename, gcs=None, cache_dir='./cache'):
        
        if len(self.steps) == 0:
            return
        
        pipeline = self.build(pipelines=[name for name in self.step_names])
        
        if 'gs://' in filename:
            if gcs is None:
                gcs = GoogleCloudStorage()
            cache_file = cache_dir + '/' + filename.split('/')[-1]
            #print(cache_file)
            joblib.dump(pipeline, cache_file)
            gcs.upload_file(cache_file, filename)
        else:
            joblib.dump(pipeline, filename)
        
        return
    


# -*- coding: utf-8 -*-
""" transformers.py

A collection of often-used sklearn custom Transformers.

"""

import pandas as pd
import numpy as np
import inspect, sys
import datetime

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import FeatureUnion, _fit_transform_one, _transform_one
from sklearn.externals.joblib import Parallel, delayed

from scipy import sparse

import humailib.utils as hu


def get_all_transformers():
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj):
            print(obj)


class DataframeFunctionTransformer(BaseEstimator, TransformerMixin):
    """
    Transform dataframe with a custom function.
    The function, func, needs to be of the format:
      func(X, params)
    """
    def __init__(self, func, params):
        self.func = func
        self.params = params

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None, copy=None):
        
        #assert isinstance(input_df, pd.DataFrame) or isinstance(input_df, pd.Series)
        
        print("[DataframeFunctionTransformer] with function '{}' on {:,} rows...".format(self.func.__name__, len(X)))
        print("  params={}".format(self.params))
        return self.func(X, self.params)
    
    
class DataframeFeatureUnion(FeatureUnion):
    """
    FeatureUnion that works and preserves pandas Dataframes.
    From: https://zablo.net/blog/post/pandas-dataframe-in-scikit-learn-feature-union/
    
    If using PipelineBuilder, treat a DataframeFeatureUnion as a transformer step in the
    pipeline.
    
    Example:
    
        pipeline_builder = PipelineBuilder()
        pipeline_builder.new_pipeline('aggregate_and_finalise')        
        ...
        
        df_aggregate = pipeline_builder.transform_and_add([
            ('aggregate_customers', ht.DataframeFeatureUnion([
                ('agg_max', ht.Aggregator(
                    agg_type='max', agg_column='Customer_ID', columns=['A','B'],
                    copy_agg_column=True
                )), 
                ('agg_mean', ht.Aggregator(
                    agg_type='mean', agg_column='Customer_ID', columns=['C','D'],
                    copy_agg_column=False
                )),
                ('agg_sum', ht.Aggregator(
                    agg_type='sum', agg_column='Customer_ID', columns=['E','F'],
                    copy_agg_column=False
                )) 
            ], n_jobs=3)
            )
        ], df)

    """
    
    def fit_transform(self, X, y=None, **fit_params):
        self._validate_transformers()
        result = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_transform_one)(trans, weight, X, y,
                                        **fit_params)
            for name, trans, weight in self._iter()
        )

        if not result:
            # All transformers are None
            return np.zeros((X.shape[0], 0))
        
        Xs, transformers = zip(*result)
        self._update_transformer_list(transformers)
        if any(sparse.issparse(f) for f in Xs):
            Xs = sparse.hstack(Xs).tocsr()
        else:
            Xs = self.merge_dataframes_by_column(Xs)
            
        return Xs

    def merge_dataframes_by_column(self, Xs):
        return pd.concat(Xs, axis="columns", copy=False)

    def transform(self, X):
        Xs = Parallel(n_jobs=self.n_jobs)(
            delayed(_transform_one)(trans, weight, X)
            for name, trans, weight in self._iter()
        )
        
        if not Xs:
            # All transformers are None
            return np.zeros((X.shape[0], 0))
        
        if any(sparse.issparse(f) for f in Xs):
            Xs = sparse.hstack(Xs).tocsr()
        else:
            Xs = self.merge_dataframes_by_column(Xs)
            
        return Xs
    

class CrossReferenceTransformer(BaseEstimator, TransformerMixin):
    """
    Cross reference a column in X with the same column in a reference
    dataframe, and return the values from the return column.
    """
    def __init__(self, df_ref, ref_column, ret_column, na_value=np.nan):
        self.df_ref = df_ref
        self.ref_column = ref_column
        self.ret_column = ret_column
        self.na_value = na_value
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[CrossReferenceTransformer] transform on {:,}...".format(len(X)))
        print("  Cross-referencing {}, and returning {}".format(self.ref_column, self.ret_column))
        
        a_to_b = {a:b for a,b in zip(self.df_ref[self.ref_column].values, self.df_ref[self.ret_column].values)}
    
        df = X.copy()
        df.loc[:,self.ret_column] = X.apply(
            lambda x: a_to_b.get(x[self.ref_column], self.na_value), 
            axis=1
        )
        
        return df
    

class KeepMostRecentTransformer(BaseEstimator, TransformerMixin):
    """
    Keep the most recent n rows for each entity of [agg_column]
    """
    def __init__(self, agg_column, datetime_column, n=5):
        self.agg_column = agg_column
        self.datetime_column = datetime_column
        self.n = n
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[KeepMostRecentTransformer] transform on {:,}...".format(len(X)))
        print("  agg_column={}, datetime_column={}, n={}".format(self.agg_column, self.datetime_column, self.n))
        
        df = X.copy()

        df.loc[:,'transaction'] = df.sort_values(
            by=[self.agg_column, self.datetime_column]
        ).groupby(by=self.agg_column)[self.datetime_column].rank(method='first', ascending=False)

        df = df[df['transaction'] <= self.n]
        del df['transaction']

        print("[KeepMostRecentTransformer] left {:,} rows, or {:,} {}s.".format(len(df), df[self.agg_column].nunique(), self.agg_column))

        return df
    
    
class ColumnStatsTransformer(BaseEstimator, TransformerMixin):
    """
    Print out exploratory information for certain columns in a dataframe. 
    """
    def __init__(self, columns=None, max_rows=5):
        self.columns = columns
        self.max_rows = max_rows
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[ColumnStatsTransformer] stats {}...".format(self.columns))
        
        if self.columns is not None:
            for c in self.columns:
                hu.column_stats(X, c, max_rows=self.max_rows)
            
        return X 
           
        
class SelectColumnsTransformer(BaseEstimator, TransformerMixin):
    """
    Select which columns to keep, and discard the rest.
    """
    def __init__(self, columns=None):
        self.columns = columns

    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[SelectColumnsTransformer] select {}...".format(self.columns))
        cpy_df = X[self.columns].copy()
        return cpy_df


class DropColumnsTransformer(BaseEstimator, TransformerMixin):
    """
    Drop certain columns from a dataframe.
    """
    def __init__(self, columns=None):
        self.columns = columns

    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[DropColumnsTransformer] drop {}...".format(self.columns))
        
        return X.drop(columns=self.columns)
    
    
class DropDuplicatesTransformer(BaseEstimator, TransformerMixin):
    """
    """
    def __init__(self, subset=None):
        self.subset = subset

    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[DropDuplicatesTransformer] dropping duplicates of {} on {:,}...".format(self.subset, len(X)))
    
        df = X.drop_duplicates(subset=self.subset)
    
        print("[DropDuplicatesTransformer] left: {:,}".format(len(df)))
    
        return df
    
    
class DropNATransformer(BaseEstimator, TransformerMixin):
    """
    """
    def __init__(self, subset=None):
        self.subset = subset

    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[DropDuplicatesTransformer] dropping NaNs of {} on {:,}...".format(self.subset, len(X)))
    
        df = X.dropna(subset=self.subset)
    
        print("[DropDuplicatesTransformer] left: {:,}".format(len(df)))
    
        return df
    
    
"""
""" 
class IntTypeTransformer(BaseEstimator, TransformerMixin):
    
    def __init__(self, columns = None, na_replace_value=None, drop_na=False):
        self.columns = columns
        self.na_replace_value = na_replace_value
        self.drop_na = drop_na
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[IntTypeTransformer] transforming {} on {:,}...".format(self.columns, len(X)))
        print("  na_replace_value={}, drop_na={}".format(self.na_replace_value, self.drop_na))

        df = X.copy()
        hu.columns_as_int(df, columns=self.columns, na_replace_value=self.na_replace_value, drop_na=self.drop_na)

        print("[IntTypeTransformer] left: {:,}".format(len(df)))

        return df

    
class StrTypeTransformer(BaseEstimator, TransformerMixin):
    """
    """
    def __init__(self, columns = None, uppercase=False, drop_empty=False):
        self.columns = columns
        self.uppercase = uppercase
        self.drop_empty = drop_empty
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[StrTypeTransformer] transform {} on {:,}...".format(self.columns, len(X)))
        print("  uppercase={}, drop_empty={}".format(self.uppercase, self.drop_empty))

        df = X.copy()
        hu.columns_as_str(
            df, columns=self.columns, uppercase=self.uppercase, drop_empty=self.drop_empty
        )
        
        if self.drop_empty:
            print("[StrTypeTransformer] after dropping '': {:,}".format(len(df)))
        
        return df

    
class DatetimeTypeTransformer(BaseEstimator, TransformerMixin):
    """
    """
    def __init__(self, columns = None, datetime_format='%Y-%m-%d %H:%M:%S'):
        self.columns = columns
        self.datetimeformat = datetimeformat
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[DatetimeTypeTransformer] transform {} on {:,}...".format(self.columns, len(X)))
        print("  datetimeformat={}".format(self.datetimeformat))

        df = X.copy()
        hu.columns_as_datetime(df, columns=self.columns, datetime_format=self.datetimeformat)
        
        return df
    
"""
"""    
class TransactionsCleanupTransformer(BaseEstimator, TransformerMixin):
    
    def __init__(self, datetime_column = 'Datetime', tid_column = 'Transaction_ID', cid_column = 'Customer_ID', datetime_format='%Y-%m-%d %H:%M:%S'):
        self.datetime_column = datetime_column
        self.tid_column = tid_column
        self.cid_column = cid_column
        self.datetime_format = datetime_format
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[TransactionsCleanupTransformer] transform on {:,}...".format(len(X)))
        
        print("  datetime_column={}, tid_column={}, cid_column={}".format(
            self.datetime_column, self.tid_column, self.cid_column
        ))

        df = X.copy()
        hu.transactions_cleanup(
            df, 
            datetime_column = self.datetime_column, 
            tid_column = self.tid_column, 
            cid_column = self.cid_column, 
            datetime_format=self.datetime_format
        )
        
        return df
    
    
class KeepEmailActivitiesTransformer(BaseEstimator, TransformerMixin):
    """
    """
    
    def __init__(self, activities):
        self.activities = activities
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[KeepEmailActivitiesTransformer] keeping {} on {:,}...".format(self.activities, len(X)))

        df = X.copy()
        hu.email_keep_activities(
            df,
            activities=self.activities
        )
        
        print("[KeepEmailActivitiesTransformer] left {:,}".format(len(df)))
        
        return df
    

class UniqueEmailEventsTransformer(BaseEstimator, TransformerMixin):
    """
    """
    
    def __init__(self, cid_column = 'Customer_ID', activity_column = 'Activity', activity_datetime_column = 'Activity_Datetime', verbose=True):
        self.cid_column = cid_column
        self.activity_column = activity_column
        self.activity_datetime_column = activity_datetime_column
        self.verbose = verbose
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[UniqueEmailEventsTransformer] transform on {:,}...".format(len(X)))
        print("  cid_column={}, activity_column={}, activity_datetime_column={}".format(
            self.cid_column, self.activity_column, self.activity_datetime_column
        ))

        df = X.copy()
        hu.email_ensure_zero_or_one_activity_per_broadcast(
            df, 
            cid_column = self.cid_column,
            activity_column = self.activity_column,
            activity_datetime_column = self.activity_datetime_column,
            verbose=self.verbose
        )
        
        print("[UniqueEmailEventsTransformer] left {:,}".format(len(df)))
        
        return df
    
"""
"""
class GetConfigurableProductTransformer(BaseEstimator, TransformerMixin):
    
    def __init__(self, df_products, product_column_in = 'Product_ID', product_column_out = 'Configurable_Product_ID'):
        self.df_products = df_products
        self.product_column_in = product_column_in
        self.product_column_out = product_column_out
        #print(self.datetime_column)
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[GetConfigurableProductTransformer] transform on {:,}...".format(len(X)))
        print("  product_column_in={}, product_column_out={}".format(self.product_column_in, self.product_column_out))
        
        df_items = X.copy()
        
        # Replace Product IDs with the Product ID of the configurable it belongs to.
        configurable = {
            pid:cpid for pid,cpid in zip(self.df_products[self.product_column_in],self.df_products.Configurable_Product_ID)
        }
        df_items.loc[:,self.product_column_out] = df_items[self.product_column_in].apply(
            lambda x: configurable[x] if x in configurable else -2
        )
        
        print("Number of items for which:\n 1. there is no configurable product ID {:,}\n 2. the Product ID cannot be found in df_products {:,}\nTotal items {:,}".format(
            len( df_items[df_items[self.product_column_out] == -1] ), 
            len( df_items[df_items[self.product_column_out] == -2] ), 
            len( df_items )
        ))
        
        # For those products that do not have a configurable, keep the Product ID
        #df_items.loc[:,self.product_column_out] = df_items.apply(
        #    lambda x: x[self.product_column_in] if x[self.product_column_out] < 0 else x[self.product_column_out], 
        #    axis=1
        #)
        no_prod_ids = df_items[self.product_column_out] < 0
        df_items.loc[no_prod_ids, self.product_column_out] = df_items.loc[no_prod_ids, self.product_column_in]

        return df_items
    
    
class SeasonalityTransformer(BaseEstimator, TransformerMixin):
    
    """
    Calculate seasonality labels (one-hot encoding type) for each row
    
    Input dataframes could be:
        1. Items (to get product seasonality)
        2. Transactions (to get customer seasonality)
    """
    
    def __init__(self, datetime_column = 'Datetime'):
        self.datetime_column = datetime_column
        #print(self.datetime_column)
        return
    
    def fit(self, X, y=None):
        return
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[SeasonalityTransformer] transform on {:,}...".format(len(X)))
        
        df = X.copy()
        #print(df.head())
    
        print("  spring...")
        df.loc[:,'spring'] = df.apply(
            lambda x: 1.0 if (x[self.datetime_column].month >= 3 and x[self.datetime_column].month < 6) else 0.0, axis=1
        )

        print("  summer...")
        df.loc[:,'summer'] = df.apply(
            lambda x: 1.0 if (x[self.datetime_column].month >= 6 and x[self.datetime_column].month < 9) else 0.0, axis=1
        )

        print("  autumn...")
        df.loc[:,'autumn'] = df.apply(
            lambda x: 1.0 if (x[self.datetime_column].month >= 9 and x[self.datetime_column].month < 12) else 0.0, axis=1
        )

        print("  winter...")
        df.loc[:,'winter'] = df.apply(
            lambda x: 1.0 if (x[self.datetime_column].month == 12 or 
                              (x[self.datetime_column].month >= 1 and x[self.datetime_column].month < 3)) else 0.0, axis=1
        )

        return df


class GenderTransformer(BaseEstimator, TransformerMixin):
    
    """
    Calculate gender labels (one-hot encoding type) for each row
    
    Input dataframes could be:
        1. Items (to get product 'gender', i.e. what proportion of the product is bought by male, female, unisex)
    """
    
    def __init__(self, df_trans, df_customers, cid_column = 'Customer_ID', gender_column = 'Gender'):
        self.cid_to_gender = {cid:gender for cid,gender in zip(df_customers[cid_column].values, df_customers[gender_column].values) }
        self.tid_to_gender = {tid:self.cid_to_gender[cid] for tid,cid in zip(df_trans['Transaction_ID'].values, df_trans[cid_column].values)}
        self.cid_column = cid_column
        #print(self.datetime_column)
        return
    
    def fit(self, X, y=None):
        return
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[ProductGenderTransformer] transform on {:,}...".format(len(X)))
        
        df = X.copy()
        
        df.loc[:,'male'] = df['Transaction_ID'].apply(
            lambda x: 1.0 if (x in self.tid_to_gender and self.tid_to_gender[x] == 'M') else 0.0, 
                                   axis=1
        )
        df.loc[:,'female'] = df['Transaction_ID'].apply(
            lambda x: 1.0 if (x in self.tid_to_gender and self.tid_to_gender[x] == 'F') else 0.0, 
                                   axis=1
        )
        df.loc[:,'unisex'] = df['Transaction_ID'].apply(
            lambda x: 1.0 if (x in self.tid_to_gender and (self.tid_to_gender[x] != 'M' and self.tid_to_gender[x] != 'F')) else 0.0, 
                                   axis=1
        )        

        return df

    
class Aggregator(BaseEstimator, TransformerMixin):
    """
    """ 
    def __init__(self, agg_type, agg_column, columns, copy_agg_column=True):
        self.agg_type = agg_type
        self.agg_column = agg_column
        self.columns = columns
        self.copy_agg_column=copy_agg_column
        return
    
    def fit(self, X, y=None):
        return
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        func = 'mean'
        if self.agg_type == 'mean':
            func = pd.DataFrame.mean
        elif self.agg_type == 'sum':
            func = pd.DataFrame.sum
        elif self.agg_type == 'min':
            func = pd.DataFrame.min
        elif self.agg_type == 'max':
            func = pd.DataFrame.max
        
        n = X[self.agg_column].nunique()
        print("[Aggregator] transform on {:,}...".format(n))
        print("  method={}, agg_column={}, columns={}".format(func.__name__, self.agg_column, self.columns))
        
        dfg = X.sort_values(by=self.agg_column).groupby(by=self.agg_column)
        
        i = 0
        aggs = []
        for pid, df in dfg:
            
            if i % 1000 == 0:
                print("{:,}/{:,} ({:.2f} %)".format(i, n, 100.0*i/n))
            
            agg = [func(df[c]) for c in self.columns]
                
            if self.copy_agg_column:
                aggs.append(tuple([pid]+agg))
            else:
                aggs.append(tuple(agg))
            
            i = 1 + i
        
        if self.copy_agg_column:
            columns = [self.agg_column] + self.columns
        else:
            columns = self.columns
            
        df_out = pd.DataFrame(aggs, columns=columns)

        return df_out
    
    
class AggregatorMean(Aggregator):
    """
    """ 
    def __init__(self, agg_column, columns, copy_agg_column=True):
        super.__init__('mean', agg_column, columns, copy_agg_column)
        return
    
        
class OrderSegmentTransformer(BaseEstimator, TransformerMixin):
    """
    """        
    
    def __init__(self, num_trans_column = 'n_trans'):
        self.num_trans_column = num_trans_column
        return
    
    def fit(self, X, y=None):
        return
    
    def __get_segment(self, x):
        if x.n_trans >= 3:
            return '3+'
        if x.n_trans == 2:
            return '2'
        return '1'

    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[OrderSegmentTransformer] transform on {:,}...".format(len(X)))

        df = X.copy()
        
        df.loc[:,'order_segment'] = df.apply(lambda x: self.__get_segment(x), axis=1)
        
        return df
    
    



import sys
import pandas as pd

import requests
import rauth
from rauth import OAuth2Service
import json

class MailchimpREST:
    
    def __init__(self, access_token = None, client_id = '725681533252', client_secret = '68e3936817108f1e9ce8fe7c4021a05a8d0cffddb5e3d82c0a'):
        
        self.access_token = access_token
        
        self.oauth_base_uri = 'https://login.mailchimp.com/oauth2/'
        self.authorize_uri = 'https://login.mailchimp.com/oauth2/authorize'
        self.access_token_uri = 'https://login.mailchimp.com/oauth2/token'
        self.base_uri = ''

        self.client_id=client_id
        self.client_secret=client_secret
        self.redirect_uri='http://127.0.0.1'
        
        
    def get_auth_url(self, client_id = None, client_secret = None):
        
        if client_id is not None:
            self.client_id = client_id
            
        if client_secret is not None:
            self.client_secret = client_secret
        
        oauth2 = OAuth2Service(
            client_id=self.client_id,
            client_secret=self.client_secret,
            name='Mailchimp',
            authorize_url=self.authorize_uri,
            access_token_url=self.access_token_uri,
            base_url=self.base_uri)

        params = {
            'response_type': 'code',
            'redirect_uri': self.redirect_uri
        }

        url = oauth2.get_authorize_url(**params)
        return url
    
    def get_access_token(self, code):
        
        data={
            'grant_type':'authorization_code',
            'client_id' : client_id,
            'client_secret' : client_secret,
            'redirect_uri' : redirect_uri,
            'code' : code
        }
        
        response = requests.post(access_token_uri, data=data)
        if response.status_code == 200:
            
            self.access_token = response.json()['access_token']
            print("[get_access_token] Success. Token = '{}'".format(self.access_token))
            
        else:
            print("[get_access_token] Failed.")
            print(response.json())
            
        return self.access_token
            
        #print(response.json())
        
    def _request(self, request_func, endpoint, params={}, data={}, json={}, verbose=False):
        
        if self.access_token == None or self.access_token == '':
            print("[_request] Access token '{}' is not valid.".format(self.access_token))
            return None
        
        if self.base_uri == '':
            response = requests.get(self.oauth_base_uri + 'metadata', headers={
                'Authorization':'OAuth {}'.format(self.access_token)
            })
            if response.status_code == 200:
                #print(response.json())
                self.base_uri = response.json()['api_endpoint'] + '/3.0/'                
            else:
                print("[_request] Could not retrieve metadata.")
                return None
        
        url = self.base_uri + endpoint
        response = request_func(url, params=params, data=data, json=json, headers={
            'Authorization':'OAuth {}'.format(self.access_token)
        })

        if verbose:
            print("REST URI: {}".format(url))
            print("Response code: {0}".format(response.status_code))
            print("Response: " + response.text)

        return response, response.json()
            
        
    def post(self, endpoint, params={}, data={}, json={}, verbose=False):

        return self._request(requests.post, endpoint, params=params, data=data, json=json, verbose=verbose)

    
    def delete(self, endpoint, verbose=False):
        
        return self._request(requests.delete, endpoint, verbose=verbose)


    def put(self, endpoint, params={}, data={}, json={}, verbose=False):
        
        return self._request(requests.put, endpoint, params=params, data=data, json=json, verbose=verbose)

    
    def patch(self, endpoint, params={}, data={}, json={}, verbose=False):
        
        return self._request(requests.patch, endpoint, params=params, data=data, json=json, verbose=verbose)

    
    def get(self, endpoint, params={}, verbose=False):
        
        return self._request(requests.get, endpoint, params=params, verbose=verbose)


class MailchimpCustomersExtract():
    
    """
    Wrapper class for extracting customers from a Mailchimp audience.
    """
    
    def __init__(self, access_token = '1976af7eb0470a7d844ec590697f37d5'):
        
        self.api = MailchimpREST(access_token)
        self.df_customers = None
        self.list_id = ''
        
    def _extract_field(self, customer, field_name):
        
        return customer[field_name] if field_name in customer else ''
    
    def _extract_array(self, customer, field_name):
        
        return customer[field_name] if field_name in customer else ['']
    
    def _extract_dict(self, customer, field_name):
        
        return customer[field_name] if field_name in customer else {}
    
    def _get_audience_list_id(self, audience='HUMAI'):
        
        response, data = self.api.get('lists', verbose=False)
        for l in data['lists']:
            if l['name'] == audience:
                self.list_id = l['id']
                return True
                
        return False
    
    def extract(self, audience='HUMAI', page_size = 1000, verbose=False):
        
        if not self._get_audience_list_id(audience=audience):
            print("[extract] Could not find list ''".format(audience))
            return None
    
        offset = 0
        total_members = 1
        page_size = 1000
        customers = []
        while offset < total_members:

            try:
                response, data = self.api.get('lists/{}/members'.format(self.list_id), params={
                    'count' : page_size,
                    'offset' : offset,

                    'fields' : 'merge_fields,total_items,members'
                })

                total_members = data['total_items']
                if offset % 5000 == 0:
                    print("{:,}/{:,} ({:.2f}%)".format(offset, total_members, 100.0*offset/total_members))

                if response.status_code == 200:

                    for member in data['members']:

                        fields = self._extract_dict(member, 'merge_fields')
                        address = self._extract_dict(fields, 'ADDRESS')
                        if isinstance(address, str):
                            address = {}
                            
                        customers.append( 
                            (
                                self._extract_field(member, 'id'),
                                (self._extract_field(fields,'FNAME') + ' ' + self._extract_field(fields,'LNAME')).strip(),
                                self._extract_field(member, 'email_address').strip(),
                                self._extract_field(fields, 'BIRTHDAY'),
                                self._extract_field(address, 'addr1').strip().upper(),
                                self._extract_field(address, 'city').strip().upper(),
                                self._extract_field(address, 'zip').strip().upper(),
                                self._extract_field(address, 'country').strip().upper()
                            )
                        )
                else:
                    print("[extract] Got response status {} ({}), skipping to next offset {:,}".format(response.status_code, response.text, offset+page_size))
            except:
                print("[extract] Unexpected error: " + sys.exc_info()[0])


            offset = page_size + offset

        print("{:,}/{:,} ({:.2f}%)".format(total_members, total_members, 100.0))
        
        self.df_customers = pd.DataFrame(customers, columns=[
            'Customer_ID', 'Name', 'Email', 'Birthday', 'Address', 'City', 'Zip', 'Country' 
        ])

        return self.df_customers


        
class MailchimpPredictionsLoader:
    
    """
    Helper class for pushing predictions to Mailchimp.
    
    TODO: Finish implementation.
    """
    
    def __init__(self, audience='HUMAI', access_token = '1976af7eb0470a7d844ec590697f37d5'):
        
        self.api = MailchimpREST(access_token)
        self.ready = False
        self.audience = audience
        self.list_id = ''
        self.field_to_tag = {}
        
    def create_or_init(self, audience='HUMAI'):
        
        self.audience = audience
        self.field_to_tag = {}
        self.list_id = ''
        self.ready = False
        self._create_fields(audience)
        
        
    def _get_audience_list_id(self, audience='HUMAI'):
        
        response, data = self.api.get('lists', verbose=False)
        for l in data['lists']:
            if l['name'] == audience:
                self.list_id = l['id']
                return True
                
        return False
    
        
    def _create_fields(self, audience):
        
        try:
        
            self.ready = False
            if not self._get_audience_list_id(audience):
                print("[_create_fields] Could not find audience ''".format(audience))
                return self.ready

            required_fields = [
                ('Dates', 'date', True, ''),
                ('Time of day', 'text', True, '')
            ]

            response, data = self.api.get('lists/{}/merge-fields'.format(self.list_id))

            self.field_to_tag = {e['name']:e['tag'] for e in data['merge_fields'] }
            #print(self.field_to_tag.keys())
            
            for field in required_fields:
                if field[0] not in self.field_to_tag:
                    
                    print(field[0])
                    
                    response = self.api.post('lists/{}/merge-fields'.format(self.list_id), json={
                        'name' : field[0],
                        'type' : field[1],
                        'public' : field[2]
                    }, verbose=False)
                    
                    data = response.json()
                    #print(data)
                    
                    self.field_to_tag[field[0]] = data['tag']
        
            self.ready = True
            
            #print(self.field_to_tag)
            
        except:
            
            print("[extract] Unexpected error: " + sys.exc_info()[0])
            
        return self.ready
        
    
    def push(self, df, verbose=False):

        if not self.ready:
            print("[push] Not ready, call create_or_init()")
            return False
        
        nn = len(df)
        ii = 0

        dfg = df.groupby(by='Customer_ID')
        for cid, df_predictions in dfg:
            
            if ii % 500 == 0:
                print("{:,}/{:,} ({:.2f} %)".format(ii, nn, 100.0*ii/nn))
            
            if cid == '':
                print("  Skipping invalid Customer ID: {} with predictions {}".format(copernica_id, df_predictions))
                ii = len(df_predictions) + ii
            else:
                
                for index, row in df_predictions.iterrows():
                    
                    #print(row)
                    
                    response = self.api.patch('lists/{}/members/{}'.format(self.list_id, cid), json={
                        'merge_fields' : {
                            self.field_to_tag['Dates'] : row['Date'].strftime('%Y-%m-%d'),
                            self.field_to_tag['Time of day'] : row['Time_Of_Day']
                        }
                    }, verbose=False)


            ii = 1 + ii
            
        print("{:,}/{:,} ({:.2f} %)".format(nn, nn, 100.0))
        
        return True
        

    def remove(self, df, pred_type = 'PersonalizedMailFlow', verbose=False):

        if not self.ready:
            print("[push_predictions_to_copernica] Not ready, call create_or_init()")
            return False
        
        ## TODO: Implementation
        
        return False 
    